/* General socket routines for rwhoisd
 * specificly connecting to remote referal
 * servers
 * Note: Need to integrate arena memory to
 * this module. --a
 */

#include <stdio.h>
#include <stdlib.h>
#include <str.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <setjmp.h>
#include <fcntl.h>
#include <sys/time.h>
#include "sock_calls.h"
#include "log.h"
#include "client_msgs.h"

#define MAXREAD 4096

extern int errno;
static int time_out_flag = 0;
static jmp_buf jmpbuf;

/* funcs pre-defined */

static void
connect_timeout(int signo)
{
    longjmp(jmpbuf, 1);
}

void
fatality(char *s)
{
  fprintf(stderr,"Error: %s\n", s);
  exit(EXIT_FAILURE);
}

void
say(char *s)
{
  log(L_LOG_INFO, LOG, "[debug] %s", s);
  fflush(NULL);
}


/* b/c everyone is different */
void
con_cat(char *s1, char *s2)
{
  s1 += strlen(s1);
  while ((*s1++ = *s2++))
    ;
}


int
connect_wtimeout(int sockfd, const struct  sockaddr *saptr, socklen_t salen, int nsec)
{
	int				flags, n, error;
	socklen_t		len;
	fd_set			rset, wset;
	struct timeval	tval;

	flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

	error = 0;
	if ( (n = connect(sockfd, (struct sockaddr *) saptr, salen)) < 0)
		if (errno != EINPROGRESS)
			return(-1);

	/* Do whatever we want while the connect is taking place. */

	if (n == 0)
		goto done;	/* connect completed immediately */

	FD_ZERO(&rset);
	FD_SET(sockfd, &rset);
	wset = rset;
	tval.tv_sec = nsec;
	tval.tv_usec = 0;

	if ( (n = select(sockfd+1, &rset, &wset, NULL,
					 nsec ? &tval : NULL)) == 0) {
		close(sockfd);		/* timeout */
		errno = ETIMEDOUT;
		return(-1);
	}

	if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset)) {
		len = sizeof(error);
		if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0)
			return(-1);			/* Solaris pending error */
	} else
		say("select error: sockfd not set");

done:
	fcntl(sockfd, F_SETFL, flags);	/* restore file status flags */

	if (error) {
		close(sockfd);		/* just in case */
		errno = error;
		return(-1);
	}
	return(0);
}


/* ripped from stevens 1st ed. */
int
ref_readline(int fd, char *ptr, int maxlen)
{
  int n, rc;
  char c;

 
  for (n = 1; n < maxlen; n++) {
      if ((rc = read(fd, &c, 1)) == 1) {
	*ptr++ = c;
	if (c == '\n')
	  break;
	if (c == '%')
	  continue;
      } else if (rc == 0) {
	if (n ==1)
	  return 0;
	else
	  break;
      } else
	return(-1);
  }
  *ptr = 0;
  return(n);
}
	
void
initconn(int fd, char *data)
{
    int numbytes;
    fd_set read_set, write_set;
    struct timeval timer;
    int numsecs = 5;
    int  n;
    char readbuff[MAXREAD];
    char readbuff2[MAXREAD];
    char *response_buff;
    char *cleaned;

    response_buff = (char *)xmalloc(sizeof(char)*MAXREAD);

    /* set the read and write sets, we really
     * only have the read set for this one ;)
     */ 
    FD_ZERO(&read_set);
    FD_SET(fd, &read_set);
    write_set = read_set;
   
    /* set the timer */
    timer.tv_sec = numsecs;
    timer.tv_usec = 0;

    /* Need to move this down to right after
     * the connect. That way we could determine
     * whether we are reading or writing first.
     * Like, duh! Keep it here for now to test
     * it first ;) 
     * 
     * Now check to see if readable w/ in
     * 5 seconds 
     */
    if ((n = select(fd + 1, &read_set, NULL, NULL,
                        numsecs ? &timer : NULL)) == 0) {
        close(fd);
	    log(L_LOG_INFO, LOG, "select called timeout" );
        errno = ETIMEDOUT;
        free(response_buff);
        return;
    }

    /* Check for read/writeability baby! */
    if ( FD_ISSET(fd, &read_set)) {
     /* Now read from the socket */
	n = ref_readline(fd, readbuff, MAXREAD);
        /* the conn does not close yet */
    } else {
      say("cannot read the data");
        close(fd);
        free(response_buff);
        return;
    }

    /* finished reading the initial string version */

  if (numbytes < 0)
    say("Negative bytes read, should be instant death");

  fflush(NULL);
  
  if ( FD_ISSET(fd, &write_set)) {
    write(fd, data, strlen(data)); 
  } else {
    say("error in the write portion..");
  }

  /* set to read it back */

  if ( FD_ISSET(fd, &read_set)) {
     /* Now read from the socket */
	while((n = ref_readline(fd, readbuff2, MAXREAD)) > 0) {
        readbuff2[n] = 0;
	 /* b/c stdout is now the client ... */
      printf("%s", readbuff2);
	}
	
    } else {
        say("cannot read the msg from the remote server");
        close(fd);
    }

  free(response_buff);

  /* Fall through .. we do not have to rset the timer */
    return;
}


int
gen_attach(char *host, int port, char *data)
{

  int sockfd, nsec, n;
  int *sigfunc;
  char *err_string;
  struct sockaddr_in servaddr;
  struct hostent *hp;
  struct in_addr inaddr;

  /* Not sure about how much time here for
   * a default, 10 seconds?
   */
  nsec = 10;

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
    fatality("Massive trauma establishing local socket");
  
  bzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(port);
  
  
  if (inet_aton(host, &inaddr) == 1) {
    servaddr.sin_addr = inaddr;
   } else if ((hp = gethostbyname(host)) != NULL) {
    bcopy(hp->h_addr, &servaddr.sin_addr, hp->h_length);
   } else {
    say("host does not exist!");
    return -1;
   }
   
    /* timeout doesnt work
     * if (signal(SIGALRM, connect_timeout) == SIG_ERR)
     *  say("error attempting to catch SIGALRM!!");
     *
     * alarm(10);
     */

  say("in gen_attach ... prep to connect");
  if (( n = connect_wtimeout(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr), 5)) < 0) {
    close(sockfd);
    if (errno == EINTR)
        errno = ETIMEDOUT;
    log(L_LOG_INFO, LOG, "[debug] no conn to host: <%s>", host);
    return -1; 
  }

  alarm(0);

    /* switched around for rwhois a litte
     * first we read from the port
     * second send our query string
     * read again for out result
     */
    say("done w/ connect .. prep to initconn");

  /* call initconn to read _and_ write the data */
  initconn(sockfd,data);
  
  return 1;
}


/* ref_adder should be the rwhois server to connect to
 * and ref_query should be the ip adder query string
 */
int
ref_check(char *ref_adder, char *ref_query)
{
    struct itimerval itimer;
    char *save, *save2;
	char *proto, *rwho_host, *port, *tokv, *tokz;
	int rport, rval;
	char *terminator = "\n";
    int err_flag = 0;

    /* init the interval timer */
    itimer.it_interval.tv_usec = 0;
    itimer.it_interval.tv_sec  = 0;
    itimer.it_value.tv_usec = 0;
    itimer.it_value.tv_sec  = 10;

    /* Init some of the vars */
    save = save2 = proto = NULL;

    /* ref_adder is the rwhois host in rwhois://host.foo.com:4321
     * fashion. ref_query is the query string we want to send */
    save = strdup(ref_adder);
    save2 = strdup(ref_query);
    con_cat(save2, terminator);


    /* Have to check how it matches 
     * ref_adder can have rwhois://foo.foo.com:1234/
     * or hostname.foo.com:1234:rwhois
     * or hostname.foo.com:1234
     */

     proto = (char *)xmalloc(sizeof(char)*MAXREAD);
     rwho_host = (char *)xmalloc(sizeof(char)*MAXREAD);
     port = (char *)xmalloc(sizeof(char)*MAXREAD);

    say("in ref_check .. prep to parse");
    log(L_LOG_INFO, LOG, "[debug] string to parse: <%s>", save);
    log(L_LOG_INFO, LOG, "[debug] ref_adder in ref_check: <%s>", ref_adder);
    log(L_LOG_INFO, LOG, "[debug] ref_query in ref_query: <%s>", ref_query);

    /* set up the timer */
    if (signal(SIGALRM, connect_timeout) == SIG_ERR)
       say("error attempting to set SIGALRM!!");
   
    /* setjmp to here to buffer the call to str_parse
     * something is seriously wrong with it and I haven't
     * found the problem yet :P --a
     */
    if(setjmp(jmpbuf) == 0) {

        setitimer(ITIMER_REAL, &itimer, (struct itimerval *) 0);
     /* call the parse routine */
         if (str_parse(save,"m|^(rwhois)://(.+):(\\d+).*$|", &proto, &rwho_host, &port ) == 1) { 
             rport = atoi(port);
             say("parsed matched first");
	     }  else if (str_parse(save, "m/^(.+?):(rwhois)$/", &rwho_host, &port) == 1) {
        	  rport = 4321;
              say("parse matched second");
         } else if (str_parse(save, "m/^(.+?):(\\d+?)$/", &rwho_host, &port) == 1 ) {
        	  rport = atoi(port);
              say("parse matched third");
         } else {
              say("parse failed");
              err_flag = 1;
         }
    } else {
        say("error handling str_parse :(");
        err_flag = 1;
    }

    /* disable timer */
    itimer.it_value.tv_usec = 0;
    itimer.it_value.tv_sec  = 0;
    setitimer(ITIMER_REAL, &itimer, (struct itimerval *) 0);

    /* call to actually attach socket */
    if(err_flag != 1) {
        if(gen_attach(rwho_host, rport, save2) < 0)
         err_flag = 1;
    } 
    
    say("done w/ parse and gen_conn, prep to free mem");
  /* free all the mem */
  free(proto);
  free(rwho_host);
  free(port);

  say("done w/ ref_check");

  if (err_flag == 1) {
    return -1;
  } else { 
    return 1;
  }


}
