/* *************************************************************
   RWhois Software

   Copyright (c) 1994 Scott Williamson and Mark Kosters
   Copyright (c) 1996-2000 Network Solutions, Inc.

   See the file LICENSE for conditions of use and distribution.
**************************************************************** */

#include "decrypt_msg.h"

#include "client_msgs.h"
#include "defines.h"
#include "fileutils.h"
#include "log.h"
#include "main_config.h"
#include "misc.h"
#include "strutil.h"

#define KEYID_FILE_TEMPLATE   "%s/keyid.%s.XXXXXX"


/* Local function prototypes */

static void find_keyid PROTO((char *input_file, 
                              char *output_file, 
                              char *pubring,
                              char *keyid_file));

static char *get_userid PROTO((char *instr));

static char *get_keyid PROTO((char *instr));


int
verify_pgp_key(keyid)
  char *keyid;
{
  FILE *fp;
  char  output_file[MAX_LINE];
  char  buf[MAX_LINE];
  int   rval = FALSE;

  if (!keyid || !*keyid)
  {
    log(L_LOG_ERR, UNKNOWN, "verify_pgp_key: keyid required\n");
    return(rval);
  }

  bzero((char *)output_file, sizeof(output_file));
  strcpy(output_file, "/tmp/pgp.XXXXXX");
  mktemp(output_file);
  unlink(output_file);
  
  bzero((char *)buf, sizeof(buf));
  sprintf(buf, "pgp -kv 0x%s > %s", keyid, output_file);
  system(buf);
    
  if ((fp = fopen(output_file, "r")) == NULL)
  {
    log(L_LOG_ERR, UNKNOWN,
        "verify_pgp_key: could not open file '%s': %s\n", output_file,
        strerror(errno));
    return(rval);
  }

  fseek(fp, 0L, 0);
  bzero((char *)buf, sizeof(buf));
  while (fgets(buf, MAX_LINE, fp))
  {
    rtrim(buf);
    if (!strncmp(buf, "1 matching key found", 20))
    {
      rval = TRUE;
      break;
    }
    bzero((char *)buf, sizeof(buf));
  }
  fclose(fp);

  unlink(output_file);

  return(rval);
}


int
decrypt_pgp_msg(input_file, output_file, keyid, pubring)
  char *input_file;
  char *output_file;
  char *keyid;
  char *pubring;
{
  FILE *fp;
  char  keyid_file[MAX_LINE];
  char  buf[MAX_LINE];
  int   rval = FALSE;

  if (!input_file || !*input_file)
  {
    log(L_LOG_ERR, UNKNOWN, "decrypt_pgp_msg: input file name required");
    return(rval);
  }
  if (!output_file || !*output_file)
  {
    log(L_LOG_ERR, UNKNOWN, "decrypt_pgp_msg: output file name required");
    return(rval);
  }

  unlink(output_file);

  if (create_filename(keyid_file, KEYID_FILE_TEMPLATE,
                      get_register_spool()) == NULL)
  {
    log(L_LOG_ERR, UNKNOWN, "decrypt_pgp_msg: can not create file name: %s",
        strerror(errno));
    return(rval);
  }

  bzero((char *)buf, sizeof(buf));

  find_keyid(input_file, output_file, pubring, keyid_file);

  if ((fp = fopen(keyid_file, "r")) == NULL)
  {
    log(L_LOG_ERR, UNKNOWN,
        "decrypt_pgp_msg: could not open file '%s': %s", keyid_file,
        strerror(errno));
    return(rval);
  }

  fseek(fp, 0L, 0);
  bzero((char *)buf, sizeof(buf));
  while (fgets(buf, MAX_LINE, fp))
  {
    rtrim(buf);
    if (buf[0] != '\0')
    {
      strcpy(keyid, buf);
      rval = TRUE;
      break;
    }
    bzero((char *)buf, sizeof(buf));
  }
  fclose(fp);

  unlink(keyid_file); 

  return(rval);
}



static void
find_keyid(input_file, output_file, pubring, keyid_file)
  char *input_file;
  char *output_file;
  char *pubring;
  char *keyid_file;
{
  FILE  *pgpfp1;
  FILE  *pgpfp2;
  FILE  *in;
  FILE  *out;
  FILE  *err;
  FILE  *fp;
  char  buf1[1024];
  char  buf2[1024];
  char  *userid = NULL;
  char  keyid[MAX_LINE];
  char  *key = NULL;
  int   fd[2];
  char  *cmd = "pgp +batchmode";
  int   childpid;
  char  *newcmd;
 
  if ((childpid = fork()) < 0)
  {
    exit(-1);
  }
  if (childpid == 0)
  {
    if (pipe(fd) < 0)
    {
      fprintf(stderr, "Pipe failed.");
      exit(-1);
    }

    if ((in = fdopen(fd[0], "r")) == NULL)
    {
      fprintf(stderr,"fdopen error");
      exit(1);
    }
    if ((out = fdopen(fd[1], "w")) == NULL)
    {
      fprintf(stderr,"fdopen error");
      exit(1);
    }
    if ((err = fdopen(fd[1], "w")) == NULL)
    {
      fprintf(stderr,"fdopen error");
      exit(1);
    }

    if (dup2(fd[0], 0) == -1)
    {
      fprintf(stderr,"dup error");
      exit(1);
    }
    if (dup2(fd[1], 1) == -1)
    {
      fprintf(stderr,"dup error");
      exit(1);
    }
    if (dup2(fd[1], 2) == -1)
    {
      fprintf(stderr,"dup error");
      exit(1);
    }

    /*  Now that the environment for I/O has been setup, continue with the
        processing. */
 
    if (!(newcmd = (char *)calloc(1, sizeof(char) * strlen(cmd) + 
                                  strlen(input_file) + strlen(output_file)
                                  + 10)))
    {
      
      exit (-1);
    }

    sprintf(newcmd, "%s %s -o %s", cmd, input_file, output_file);
     
    if ((pgpfp1 = popen(newcmd, "r")) == NULL)
    {
      exit(-1);
    }
    while (fgets(buf1, 1024, pgpfp1) != NULL)
    {
      if (strncmp(buf1, "Good", 4) != 0)
      {
        continue;
      }
      userid = get_userid(buf1);

      /* extract the keyID of this user */
      
      free(newcmd);
      
      if (!(newcmd = (char *)calloc(1, sizeof(char) * strlen(cmd) + 
                                    strlen(userid) + strlen(pubring) + 10)))
      {
        exit (-1);
      }

      if (!pubring || !(*pubring))
      {
        sprintf(newcmd, "%s -kv %s", cmd, userid);
      }
      else
      {
        sprintf(newcmd, "%s -kv %s %s", cmd, userid, pubring);
      }
      if ((pgpfp2 = popen(newcmd, "r")) == NULL)
      {
        exit(-1);
      }
      while (fgets(buf2, 1024, pgpfp2) != NULL)
      {
        if (strncmp(buf2, "pub", 3) != 0)
        {
          continue;
        }
        key = get_keyid(buf2);
        if(keyid[0] == '\0')
        {
          sprintf(keyid, "%s ", key);
        }
        else 
        {
          strcat(keyid, key);
          strcat(keyid, " ");
        }
      }
      pclose(pgpfp2);
    }
    pclose(pgpfp1);
    if ((fp = fopen(keyid_file, "w")) == NULL)
    {
      exit(-1);
    }
    fprintf(fp, "%s", keyid);
    
    fclose(fp);
        
    fclose(in);
    fclose(out);
    fclose(err);
    
    free(userid);
    free(key);
    free(cmd);
    free(newcmd);

    exit(0);
  }
  else 
  {
    wait((int *)NULL);
  }
}



static char * 
get_userid(instr)
  char *instr;
{
  int   i;
  int   length;
  int   start;
  char  *outstr;

  length = strlen(instr);

  for (i = 0; instr[i] != '"'; i++)
  {
  }

  if (!(outstr = (char *)calloc(1, sizeof(char) * (length - i) + 5)))
  {
    exit(-1);
  }

  outstr[0] = '"';
  start = i + 1;

  for (i = start; instr[i] != '"'; i++)
  {
    outstr[(i - start) + 1] = instr[i];
  }

  outstr[(i - start) + 1] = '"';
  outstr[(i - start) + 2] = '\0';
  
  return outstr;
}



static char * 
get_keyid(instr)
  char *instr;
{
  int   i;
  int   length;
  int   start;
  char  *keyid;

  printf("%s\n", instr);
  
  for (i = 0; instr[i] != '/'; i++)
  {
  }

  start = i + 1;

  for (i = start; instr[i] != ' '; i++)
  {
  }

  length = i - start;
  
  if (!(keyid = (char *)calloc(1, sizeof(char) * (length) + 5)))
  {
    exit(-1);
  }

  for (i = start; instr[i] != ' '; i++)
  {
    keyid[i - start] = instr[i];
  }

  keyid[i - start] = '\0';

  return keyid;
}
