/* *************************************************************
   RWhois Software

   Copyright (c) 1994 Scott Williamson and Mark Kosters
   Copyright (c) 1996-2000 Network Solutions, Inc.

   See the file LICENSE for conditions of use and distribution.
**************************************************************** */

#include "session.h"

#include "client_msgs.h"
#include "deadman.h"
#include "defines.h"
#include "directive.h"
#include "directive_conf.h"
#include "dl_list.h"
#include "dump.h"
#include "log.h"
#include "main_config.h"
#include "misc.h"
#include "parse.h"
#include "pgp_ops.h"
#include "records.h"
#include "referral.h"
#include "search.h"
#include "security_directive.h"
#include "state.h"
#include "strutil.h"

#include "conf.h"

static int processline PROTO((char *str));
static int run_query PROTO((char *str));
 
/* ------------------- LOCAL FUNCTIONS -------------------- */


static int
processline(str)
  char *str;
{
  int               status = FALSE;
    char              *storefile;
    char              *outfile;
    FILE              *outfp;
    FILE              *pwfp;
    rwhois_state_type state  = get_rwhois_state();
    FILE              *fp    = NULL;
    auth_struct       *response;
    char              *buf;
    char              passphrase[100];

  if (!str || !*str)
  {
    return TRUE;
  }

  trim(str);

  response = get_response_auth_struct();
 
  if ((get_rwhois_secure_mode() == TRUE) && response != NULL)
  {

    storefile = tempnam(".", "out");
    outfp = fopen(storefile, "w");
    set_out_fp(outfp);
    outfile = xcalloc(1, sizeof(char) * (strlen(storefile) + 10));
    sprintf(outfile, "%s.asc", storefile);
  }
  if (is_directive(str))
  {
    status = run_directive(str);

    if (status == TRUE)
    {
            
      if ((get_rwhois_secure_mode() == TRUE) && response != NULL)
      {
        fclose(outfp);
        set_out_fp(stdout);
        if (STR_EQ(response->type, "sign"))
        {
          pwfp = fopen(get_pgp_pwfile(), "r");
          fscanf(pwfp, "%s", passphrase);
          fclose(pwfp);

          PGP_sign_msg(storefile, outfile, get_pgp_uid(), passphrase); 
        }
        if (STR_EQ(response->type, "encr"))
        {
          PGP_encrypt_msg(storefile, outfile, response->info); 
        }
        if (STR_EQ(response->type, "sign-encr"))
        {
          pwfp = fopen(get_pgp_pwfile(), "r");
          fscanf(pwfp, "%s", passphrase);
          fclose(pwfp);
          
          PGP_sign_encrypt_msg(storefile, outfile, 
                               response->info, get_pgp_uid(), passphrase); 
        }
        outfp = fopen(outfile, "r");
        while(!feof(outfp) && !ferror(outfp))
        {
          buf = xcalloc(1, sizeof(char) * 1050);
          fread(buf, 1, 1024, outfp);
          fwrite(buf, 1, 1024, stdout);
          free(buf);

        }
        fclose(fp);
/*      unlink(storefile); */
/*      unlink(outfile); */
      }
      print_ok();
    }
    if (status == QUIT)
    {
      return FALSE;
    }
  }
  else
  {
    if (state == SPOOL_STATE)
    {
      fp = open_spool_file("w+");
      if (fp)
      {
        fprintf(fp, "%s", str);
        fprintf(fp, "\n");
      }
    }
    else
    {
      run_query(str);
      return(get_holdconnect());
    }
  } 

  return TRUE;
}


static int
run_query(str)
  char *str;
{
  query_struct      *query;
  dl_list_type      record_list;
  record_struct     *record;
  int               not_done;
  int               ret_code;
  int               num_hits;
  int               obj_found_flag = FALSE;
  int               ref_onoff;

  query = xcalloc(1, sizeof(*query));
  
  dl_list_default(&record_list, FALSE, destroy_record_data);
  
  if (!str || !*str)
  {
    log(L_LOG_ERR, QUERY, "run_query: null data detected");
    print_error(INVALID_QUERY_SYNTAX, "");
    return FALSE;
  }

  log(L_LOG_INFO, CLIENT, "query: %s", str);
  if (!parse_query(str, query))
  {
    log(L_LOG_INFO, CLIENT, "invalid query syntax: %s", str);
    return FALSE;
  }

  if (!check_query_complexity(query))
  {
    destroy_query(query);
    return FALSE;
  }

  /*  The actual search call is launched from here. The original code 
   *  code has this switched backwards if we wish to actually run the 
   *  referrals. We must first check to see if they are any referral 
   *  records, if we are configured to do so. And then attempt to connect 
   *  to the referral server. If any data is provided from that server  
   *  then we display that, if not we display our data 
   */

  num_hits = search(query, &record_list, get_hit_limit(), &ret_code);
  log(L_LOG_INFO, CLIENT, "query response: %d hits", num_hits);

  /* display the object results */
  if (!dl_list_empty(&record_list))
  {
    obj_found_flag = TRUE;
    
    not_done = dl_list_first(&record_list);
    while (not_done)
    {
      log(L_LOG_INFO, LOG, "[debug] supressing the dump");
      record = dl_list_value(&record_list);
      
      /* check if see if referrals are checked or not 
       * and display_dump_format accordingly
       */
      ref_onoff = isset_ref_lookup();
      if(ref_onoff == FALSE) 
      {
         log(L_LOG_INFO, LOG, "[debug] refcheck OFF");
          display_dump_format(record);
      } else {
         log(L_LOG_INFO, LOG, "[debug] refcheck ON ");
      }
      
      not_done = dl_list_next(&record_list);
    }

    dl_list_destroy(&record_list);
  }

  /* always check for referrals -- except when the query could have
     returned referral objects! Except of course, when the server is
     configured to not search for referrals */
   log(L_LOG_INFO, LOG, "[debug] prepping to check refer");

  if (!get_skip_referral_search() &&
      (NOT_STR_EXISTS(query->auth_area_name) ||
      !STR_EQ(query->auth_area_name, "Referral")))
  {
    log(L_LOG_INFO, LOG, "[debug] test to refer_query(query)");

    if (refer_query(query))
    {
      log(L_LOG_INFO, LOG, "[debug] calling refer_query");
      obj_found_flag = TRUE;
    } else {
       /* nut sure about this part as of yet */
        display_dump_format(record);
    }
   } else {
    log(L_LOG_INFO, CLIENT, "[debug] not checking refer_query");
  }

  log(L_LOG_INFO, LOG, "[debug] done checking refer");

  /* print the resulting error or ok terminator. */
  switch (ret_code)
  {
  case SEARCH_SUCCESSFUL:
    if (obj_found_flag)
    {
      print_ok();
    }
    else
    {
      print_error(NO_OBJECTS_FOUND, "");
    }
    break;
  case HIT_LIMIT_EXCEEDED:
    print_error(EXCEEDED_MAXOBJ, "");
    break;
  default:
    print_error(UNIDENT_ERROR, "");
    break;
  }   

  destroy_query(query);
  return TRUE;
}
 

/* ------------------- PUBLIC FUNCTIONS ------------------- */


/* run_session: runs a rwhois client session */
void
run_session(real_flag)
  int real_flag;
{
  char target[MAX_LINE];
  int  not_finished    = TRUE;

  randomize(); /* in case we need to generate salts, or something */

  set_out_fp(stdout);

  print_welcome_header();

  /* set the input to line buffering, not block buffering */
#ifdef SETVBUF_REVERSED
  setvbuf(stdin, _IOLBF, (char *)NULL, 0);
#else
  setvbuf(stdin, (char *)NULL, _IOLBF, 0);
#endif

  /* unbuffer the output */
#ifdef SETVBUF_REVERSED
  setvbuf(stdout, _IONBF, (char *)NULL, 0);
#else
  setvbuf(stdout, (char *)NULL, _IONBF, 0);
#endif

  if (!real_flag)
  {
    print_error(SERVICE_NOT_AVAIL, "exceeded max client sessions");

    return;
  }
  
  do
  {
    set_timer(get_deadman_time(), is_a_deadman);

    /* this is kind of a hack, but, hey, it works */
    clear_printed_error_flag();
    
    if (readline(stdin, target, MAX_LINE) == NULL)
    {
      not_finished = FALSE;
    }
    else
    {
      unset_timer();
      not_finished = processline(target);
    }    
  } while (not_finished);
}


/* print_welcome_header: prints the standard rwhois banner greeting */
void
print_welcome_header()
{
  char             *hostname;
  dl_list_type     *dir_list;
  directive_struct *dir;
  int              not_done;
  long             capid      = 0x00;


  hostname = get_local_hostname();
  
  dir_list = get_directive_list();

  not_done = dl_list_first(dir_list);
  while (not_done)
  {
    dir = dl_list_value(dir_list);
    if (!(dir->disabled_flag))
    {
      if (dir->cap_bit > 0) 
      {
        capid = capid | dir->cap_bit;
      }
    }
    not_done = dl_list_next(dir_list);
  }

  print_response(RESP_RWHOIS, 
                 "V-%s:%6.6x:00 %s (by Network Solutions, Inc. V-%s)",
                 RWHOIS_PROTOCOL_VERSION, capid, hostname,
                 RWHOIS_SERVER_VERSION);

  fflush(stdout);
}
